/*
 *  tslib/plugins/debounce.c
 *
 *  Copyright (C) 2009 TomTom Ltd
 *
 * This file is placed under the LGPL.  Please see the file
 * COPYING for more details.
 *
 * $Id:  $
 *
 * Debounce filter for touchscreens
 *
 * Problem: some touchscreens have limited resolution on pressure (even
 * just on/off). This results in a situation where medium pressure can
 * result in multiple pen down, pen up events very quickly.
 *
 * Solution: present two configurable options:
 * 
 *  postUpDeadTimeMs = time in milliseconds after processing a pen up
 * before any more touchscreen events are allowed again. Note this means
 * the user needs to lift off for postUpDeadTimeMs ms before the screen
 * will work again. This stops an unintentional double tap.
 * 
 *  downIgnoreEvents = number of events to ignore after pen down before
 * passing on the rest. If the first N events from the touchscreen are
 * unreliable, then this value can be tuned appropriately.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "tslib.h"
#include "tslib-filter.h"

#define DEBUG_TS_DEBOUNCE
#undef DEBUG_TS_DEBOUNCE

struct tslib_debounce {
    struct tslib_module_info module;
    int postUpDeadTimeUs;
    int downIgnoreEvents;
    struct ts_sample last;
    struct timeval iIgnorePenDownBefore;
    int numberDownSamples; // how many times have we seen a pen down in a row
};

static int debounce_read(struct tslib_module_info *info, struct ts_sample *samp, int nr)
{
    struct tslib_debounce *var = (struct tslib_debounce *)info;
    struct ts_sample *s;
    int count = 0, ret;

    ret = info->next->ops->read(info->next, samp, nr);
    for (s = samp; ret > 0; s++, ret--) {
	if (s->pressure > 0) { // pen down
	    if ((s->tv.tv_sec > var->iIgnorePenDownBefore.tv_sec ||
		 (s->tv.tv_sec == var->iIgnorePenDownBefore.tv_sec &&
		  s->tv.tv_usec > var->iIgnorePenDownBefore.tv_usec))) { 
		// event not being debounced..
		// check whether enough events registered..
		var->numberDownSamples++;
		if (var->numberDownSamples <= var->downIgnoreEvents) {
#ifdef DEBUG_TS_DEBOUNCE
		    fprintf(stderr,
			    "%ld.%06ld : DEBOUNCE Not enough down samples: Pressure: %01d, X: %03d, Y: %03d\n",
			    s->tv.tv_sec, s->tv.tv_usec, s->pressure, s->x, s->y);
#endif
		    continue;
		}
		// enough pendowns already happened, allow the event through...
	    }
	    else {
		// suppress this event - in debounce mode, and reset debounce timer
#ifdef DEBUG_TS_DEBOUNCE
		fprintf(stderr,
			"%ld.%06ld : DEBOUNCE SUPPRESSED ReadEvent: Pressure: %01d, X: %03d, Y: %03d\n",
			s->tv.tv_sec, s->tv.tv_usec, s->pressure, s->x, s->y);
#endif
		struct timeval debounceUpDelay;
		debounceUpDelay.tv_sec = 0;
		debounceUpDelay.tv_usec = var->postUpDeadTimeUs;
		timeradd(&s->tv, &debounceUpDelay, &var->iIgnorePenDownBefore);
		continue;
	    }
	} // pressure == 0, pen up
	else {
	    var->numberDownSamples = 0;
	    // pen up - allow out, and set out debounce timer, unless we haven't just sent a down.
	    if (var->last.pressure > 0) {
		struct timeval debounceUpDelay;
		debounceUpDelay.tv_sec = 0;
		debounceUpDelay.tv_usec = var->postUpDeadTimeUs;
		timeradd(&s->tv, &debounceUpDelay, &var->iIgnorePenDownBefore);
#ifdef DEBUG_TS_DEBOUNCE
		fprintf(stderr,
			"%ld.%06ld : DEBOUNCE PenUP: Pressure: %01d, X: %03d, Y: %03d\n",
			s->tv.tv_sec, s->tv.tv_usec, s->pressure, s->x, s->y);
#endif
	    }
	    else {
#ifdef DEBUG_TS_DEBOUNCE
		fprintf(stderr,
			"%ld.%06ld : DEBOUNCE SUPPRESSED PenUP: Pressure: %01d, X: %03d, Y: %03d\n",
			s->tv.tv_sec, s->tv.tv_usec, s->pressure, s->x, s->y);
#endif
		continue;
	    }
	}
	samp [count++] = *s;
	var->last = *s;
    }
    return count;
}

static int debounce_fini(struct tslib_module_info *info)
{
    free(info);
    return 0;
}

static const struct tslib_ops debounce_ops =
{
    .read	= debounce_read,
    .fini	= debounce_fini,
};

static int debounce_limit(struct tslib_module_info *inf, char *str, void *data)
{
    struct tslib_debounce *var = (struct tslib_debounce *)inf;
    unsigned long v;
    int err = errno;

    v = strtoul(str, NULL, 0);

    if (v == ULONG_MAX && errno == ERANGE)
	return -1;

    errno = err;
    switch ((int)data) {
    case 1:
#ifdef DEBUG_TS_DEBOUNCE
	fprintf(stderr, "DEBOUNCE Setting upIgnoreTimeMs = %d\n", v);
#endif
	var->postUpDeadTimeUs = v;
	break;
    case 2:
#ifdef DEBUG_TS_DEBOUNCE
	fprintf(stderr, "DEBOUNCE Setting downIgnoreEvents = %d\n", v);
#endif
	var->downIgnoreEvents = v;
	break;

    default:
	return -1;
    }
    return 0;
}

static const struct tslib_vars debounce_vars[] =
{
    { "postUpDeadTimeMs",	(void *)1, debounce_limit }, // user supplies millis, we use microseconds
    { "downIgnoreEvents",	(void *)2, debounce_limit },
};

#define NR_VARS (sizeof(debounce_vars) / sizeof(debounce_vars[0]))

TSAPI struct tslib_module_info *mod_init(struct tsdev *dev, const char *params)
{
    struct tslib_debounce *var;

    var = malloc(sizeof(struct tslib_debounce));
    if (var == NULL)
	return NULL;

    var->module.ops = &debounce_ops;

    var->postUpDeadTimeUs = -1; // by default off
    var->downIgnoreEvents = 0; // by default off
    var->iIgnorePenDownBefore.tv_sec = 0;
    var->iIgnorePenDownBefore.tv_usec = 0;
    var->numberDownSamples = 0;
    var->last.pressure = 1;

    if (tslib_parse_vars(&var->module, debounce_vars, NR_VARS, params)) {
	free(var);
	return NULL;
    }
    var->postUpDeadTimeUs *= 1000;

    return &var->module;
}
